# -*- coding: utf-8 -*-
from __future__ import absolute_import, print_function, division
import logging
import os
import warnings

import pytest

import petl as etl
from petl.test.helpers import ieq


logger = logging.getLogger(__name__)
debug = logger.debug


def get_from_env(prefix, name, current, default):
    if current != default:
        return current
    varname = "{}_{}".format(prefix.upper(), name.upper())
    return os.getenv(varname, default)


def get_db_args(*prefixes):
    host = '127.0.0.1'
    user = 'petl'
    pwrd = 'test'
    base = 'petl'
    for prefix in prefixes:
        host = get_from_env(prefix, 'HOST', host, '127.0.0.1')
        user = get_from_env(prefix, 'USER', user, 'petl')
        pwrd = get_from_env(prefix, 'PASSWORD', pwrd, 'test')
        base = get_from_env(prefix, 'DATABASE', base, 'petl')
    return host, user, pwrd, base


def get_mysql_args(*prefixes):
    prefixes += ('PYMYSQL', 'MYSQL',)
    return get_db_args(*prefixes)


def get_pg_args(*prefixes):
    prefixes += ('PSYCOPG2', 'POSTGRESQL', 'POSTGRES', 'PG',)
    return get_db_args(*prefixes)


def _test_dbo(write_dbo, read_dbo=None):
    if read_dbo is None:
        read_dbo = write_dbo

    expect_empty = (('foo', 'bar'),)
    expect = (('foo', 'bar'),
              ('a', 1),
              ('b', 2))
    expect_appended = (('foo', 'bar'),
                       ('a', 1),
                       ('b', 2),
                       ('a', 1),
                       ('b', 2))
    actual = etl.fromdb(read_dbo, 'SELECT * FROM test')

    debug('verify empty to start with...')
    debug(etl.look(actual))
    ieq(expect_empty, actual)

    debug('write some data and verify...')
    etl.todb(expect, write_dbo, 'test')
    debug(etl.look(actual))
    ieq(expect, actual)

    debug('append some data and verify...')
    etl.appenddb(expect, write_dbo, 'test')
    debug(etl.look(actual))
    ieq(expect_appended, actual)

    debug('overwrite and verify...')
    etl.todb(expect, write_dbo, 'test')
    debug(etl.look(actual))
    ieq(expect, actual)

    debug('cut, overwrite and verify')
    etl.todb(etl.cut(expect, 'bar', 'foo'), write_dbo, 'test')
    debug(etl.look(actual))
    ieq(expect, actual)

    debug('cut, append and verify')
    etl.appenddb(etl.cut(expect, 'bar', 'foo'), write_dbo, 'test')
    debug(etl.look(actual))
    ieq(expect_appended, actual)

    debug('try a single row')
    etl.todb(etl.head(expect, 1), write_dbo, 'test')
    debug(etl.look(actual))
    ieq(etl.head(expect, 1), actual)


def _test_with_schema(dbo, schema):

    expect = (('foo', 'bar'),
              ('a', 1),
              ('b', 2))
    expect_appended = (('foo', 'bar'),
                       ('a', 1),
                       ('b', 2),
                       ('a', 1),
                       ('b', 2))
    actual = etl.fromdb(dbo, 'SELECT * FROM test')

    print('write some data and verify...')
    etl.todb(expect, dbo, 'test', schema=schema)
    ieq(expect, actual)
    print(etl.look(actual))

    print('append some data and verify...')
    etl.appenddb(expect, dbo, 'test', schema=schema)
    ieq(expect_appended, actual)
    print(etl.look(actual))


def _test_unicode(dbo):
    expect = ((u'name', u'id'),
              (u'Արամ Խաչատրյան', 1),
              (u'Johann Strauß', 2),
              (u'Вагиф Сәмәдоғлу', 3),
              (u'章子怡', 4),
              )
    actual = etl.fromdb(dbo, 'SELECT * FROM test_unicode')

    print('write some data and verify...')
    etl.todb(expect, dbo, 'test_unicode')
    ieq(expect, actual)
    print(etl.look(actual))


def _setup_mysql(dbapi_connection):
    # setup table
    cursor = dbapi_connection.cursor()
    # deal with quote compatibility
    cursor.execute('SET SQL_MODE=ANSI_QUOTES')
    cursor.execute('DROP TABLE IF EXISTS test')
    cursor.execute('CREATE TABLE test (foo TEXT, bar INT)')
    cursor.execute('DROP TABLE IF EXISTS test_unicode')
    cursor.execute('CREATE TABLE test_unicode (name TEXT, id INT) '
                   'CHARACTER SET utf8')
    cursor.close()
    dbapi_connection.commit()


def _setup_postgresql(dbapi_connection):
    # setup table
    cursor = dbapi_connection.cursor()
    cursor.execute('DROP TABLE IF EXISTS test')
    cursor.execute('CREATE TABLE test (foo TEXT, bar INT)')
    cursor.execute('DROP TABLE IF EXISTS test_unicode')
    # assume character encoding UTF-8 already set at database level
    cursor.execute('CREATE TABLE test_unicode (name TEXT, id INT)')
    cursor.close()
    dbapi_connection.commit()


def _setup_sqlalchemy_quotes(dbapi_connection, connection_record):
    cursor = dbapi_connection.cursor()
    cursor.execute("SET sql_mode = 'ANSI_QUOTES'")


SKIP_PYMYSQL = False
try:
    import pymysql
    import sqlalchemy
    myhost, myuser, mypassword, mydatabase = get_mysql_args()
    pymysql.connect(host=myhost,
                    user=myuser,
                    password=mypassword,
                    database=mydatabase)
except Exception as e:
    SKIP_PYMYSQL = 'SKIP pymysql tests: %s' % e
    warnings.warn(SKIP_PYMYSQL, UserWarning)
finally:
    @pytest.mark.skipif(bool(SKIP_PYMYSQL), reason=str(SKIP_PYMYSQL))
    def test_pymysql():

        import pymysql

        # assume database already created
        dbapi_connection = pymysql.connect(host=myhost,
                                   user=myuser,
                                   password=mypassword,
                                   database=mydatabase)

        # exercise using a dbapi_connection
        _setup_mysql(dbapi_connection)
        _test_dbo(dbapi_connection)

        # exercise using a dbapi_cursor
        _setup_mysql(dbapi_connection)
        dbapi_cursor = dbapi_connection.cursor()
        _test_dbo(dbapi_cursor)
        dbapi_cursor.close()

        # exercise sqlalchemy dbapi_connection
        _setup_mysql(dbapi_connection)
        from sqlalchemy import create_engine
        sqlalchemy_engine = create_engine('mysql+pymysql://%s:%s@%s/%s' %
                                         (myuser, mypassword, myhost, mydatabase))
        from sqlalchemy.event import listen
        listen(sqlalchemy_engine, "connect", _setup_sqlalchemy_quotes)
        sqlalchemy_connection = sqlalchemy_engine.connect()
        _test_dbo(sqlalchemy_connection)
        sqlalchemy_connection.close()

        # exercise sqlalchemy session
        _setup_mysql(dbapi_connection)
        from sqlalchemy.orm import sessionmaker
        Session = sessionmaker(bind=sqlalchemy_engine)
        sqlalchemy_session = Session()
        _test_dbo(sqlalchemy_session)
        sqlalchemy_session.close()

        # exercise sqlalchemy engine
        _setup_mysql(dbapi_connection)
        sqlalchemy_engine2 = create_engine('mysql+pymysql://%s:%s@%s/%s' %
                                          (myuser, mypassword, myhost, mydatabase), echo_pool='debug')
        listen(sqlalchemy_engine2, "connect", _setup_sqlalchemy_quotes)
        _test_dbo(sqlalchemy_engine2)
        sqlalchemy_engine2.dispose()

        # other exercises
        _test_with_schema(dbapi_connection, mydatabase)
        utf8_connection = pymysql.connect(host=myhost,
                                  user=myuser,
                                  password=mypassword,
                                  database=mydatabase,
                                  charset='utf8')
        utf8_connection.cursor().execute('SET SQL_MODE=ANSI_QUOTES')
        _test_unicode(utf8_connection)
        utf8_connection.close()


SKIP_MYSQLDB = False
try:
    import MySQLdb
    import sqlalchemy
    myhost, myuser, mypassword, mydatabase = get_db_args('MYSQLDB', 'MYSQL')
    MySQLdb.connect(host=myhost,
                    user=myuser,
                    passwd=mypassword,
                    db=mydatabase)
except Exception as e:
    SKIP_MYSQLDB = 'SKIP MySQLdb tests: %s' % e
    warnings.warn(SKIP_MYSQLDB, UserWarning)
finally:
    @pytest.mark.skipif(bool(SKIP_MYSQLDB), reason=str(SKIP_MYSQLDB))
    def test_mysqldb():

        import MySQLdb

        # assume database already created
        dbapi_connection = MySQLdb.connect(host=myhost,
                                   user=myuser,
                                   passwd=mypassword,
                                   db=mydatabase)

        # exercise using a dbapi_connection
        _setup_mysql(dbapi_connection)
        _test_dbo(dbapi_connection)

        # exercise using a dbapi_cursor
        _setup_mysql(dbapi_connection)
        dbapi_cursor = dbapi_connection.cursor()
        _test_dbo(dbapi_cursor)
        dbapi_cursor.close()

        # exercise sqlalchemy dbapi_connection
        _setup_mysql(dbapi_connection)
        from sqlalchemy import create_engine
        sqlalchemy_engine = create_engine('mysql+mysqldb://%s:%s@%s/%s' %
                                         (myuser, mypassword, myhost, mydatabase))
        from sqlalchemy.event import listen
        listen(sqlalchemy_engine, "connect", _setup_sqlalchemy_quotes)
        sqlalchemy_connection = sqlalchemy_engine.connect()
        sqlalchemy_connection.execute('SET SQL_MODE=ANSI_QUOTES')
        _test_dbo(sqlalchemy_connection)
        sqlalchemy_connection.close()

        # exercise sqlalchemy session
        _setup_mysql(dbapi_connection)
        from sqlalchemy.orm import sessionmaker
        Session = sessionmaker(bind=sqlalchemy_engine)
        sqlalchemy_session = Session()
        _test_dbo(sqlalchemy_session)
        sqlalchemy_session.close()

        # other exercises
        _test_with_schema(dbapi_connection, mydatabase)
        utf8_connection = MySQLdb.connect(host=myhost,
                                  user=myuser,
                                  passwd=mypassword,
                                  db=mydatabase,
                                  charset='utf8')
        utf8_connection.cursor().execute('SET SQL_MODE=ANSI_QUOTES')
        _test_unicode(utf8_connection)

SKIP_TEST_POSTGRES = False
try:
    import psycopg2
    import sqlalchemy
    pghost, pguser, pgpassword, pgdatabase = get_pg_args()
    psycopg2.connect(
        'host=%s dbname=%s user=%s password=%s'
        % (pghost, pgdatabase, pguser, pgpassword)
    )
except Exception as e:
    SKIP_TEST_POSTGRES = 'SKIP psycopg2 tests: %s' % e
    warnings.warn(SKIP_TEST_POSTGRES, UserWarning)
finally:
    @pytest.mark.skipif(bool(SKIP_TEST_POSTGRES), reason=str(SKIP_TEST_POSTGRES))
    def test_postgresql():

        import psycopg2
        import psycopg2.extensions
        psycopg2.extensions.register_type(psycopg2.extensions.UNICODE)
        psycopg2.extensions.register_type(psycopg2.extensions.UNICODEARRAY)

        # assume database already created
        dbapi_connection = psycopg2.connect(
            'host=%s dbname=%s user=%s password=%s'
            % (pghost, pgdatabase, pguser, pgpassword)
        )

        # exercise using a dbapi_connection
        _setup_postgresql(dbapi_connection)
        _test_dbo(dbapi_connection)

        # exercise using a dbapi_cursor
        _setup_postgresql(dbapi_connection)
        dbapi_cursor = dbapi_connection.cursor()
        _test_dbo(dbapi_cursor)
        dbapi_cursor.close()

        # exercise sqlalchemy dbapi_connection
        _setup_postgresql(dbapi_connection)
        from sqlalchemy import create_engine
        sqlalchemy_engine = create_engine('postgresql+psycopg2://%s:%s@%s/%s' %
                                          (pguser, pgpassword, pghost, pgdatabase))
        sqlalchemy_connection = sqlalchemy_engine.connect()
        _test_dbo(sqlalchemy_connection)
        sqlalchemy_connection.close()

        # exercise sqlalchemy session
        _setup_postgresql(dbapi_connection)
        from sqlalchemy.orm import sessionmaker
        Session = sessionmaker(bind=sqlalchemy_engine)
        sqlalchemy_session = Session()
        _test_dbo(sqlalchemy_session)
        sqlalchemy_session.close()

        # other exercises
        _test_dbo(dbapi_connection,
                  lambda: dbapi_connection.cursor(name='arbitrary'))
        _test_with_schema(dbapi_connection, 'public')
        _test_unicode(dbapi_connection)



# -*- coding: utf-8 -*-
from __future__ import absolute_import, print_function, division

from collections import OrderedDict
from tempfile import NamedTemporaryFile
import json

import pytest

from petl.test.helpers import ieq
from petl import dummytable, fromjson, fromdicts, tojson, tojsonarrays


def test_fromjson_1():

    f = NamedTemporaryFile(delete=False, mode='w')
    data = '[{"foo": "a", "bar": 1}, ' \
           '{"foo": "b", "bar": 2}, ' \
           '{"foo": "c", "bar": 2}]'
    f.write(data)
    f.close()

    actual = fromjson(f.name, header=['foo', 'bar'])
    expect = (('foo', 'bar'),
              ('a', 1),
              ('b', 2),
              ('c', 2))
    ieq(expect, actual)
    ieq(expect, actual)  # verify can iterate twice


def test_fromjson_2():

    f = NamedTemporaryFile(delete=False, mode='w')
    data = '[{"foo": "a", "bar": 1}, ' \
           '{"foo": "b"}, ' \
           '{"foo": "c", "bar": 2, "baz": true}]'
    f.write(data)
    f.close()

    actual = fromjson(f.name, header=['bar', 'baz', 'foo'])
    expect = (('bar', 'baz', 'foo'),
              (1, None, 'a'),
              (None, None, 'b'),
              (2, True, 'c'))
    ieq(expect, actual)
    ieq(expect, actual)  # verify can iterate twice


def test_fromjson_3():

    f = NamedTemporaryFile(delete=False, mode='w')
    data = '[{"foo": "a", "bar": 1}, ' \
           '{"foo": "b"}, ' \
           '{"foo": "c", "bar": 2, "baz": true}]'
    f.write(data)
    f.close()

    actual = fromjson(f.name, header=['foo', 'bar'])
    expect = (('foo', 'bar'),
              ('a', 1),
              ('b', None),
              ('c', 2))
    ieq(expect, actual)
    ieq(expect, actual)  # verify can iterate twice


def test_fromdicts_1():

    data = [{'foo': 'a', 'bar': 1},
            {'foo': 'b', 'bar': 2},
            {'foo': 'c', 'bar': 2}]
    actual = fromdicts(data, header=['foo', 'bar'])
    expect = (('foo', 'bar'),
              ('a', 1),
              ('b', 2),
              ('c', 2))
    ieq(expect, actual)
    ieq(expect, actual)  # verify can iterate twice


def test_fromdicts_2():

    data = [{'foo': 'a', 'bar': 1},
            {'foo': 'b'},
            {'foo': 'c', 'bar': 2, 'baz': True}]
    actual = fromdicts(data, header=['bar', 'baz', 'foo'])
    expect = (('bar', 'baz', 'foo'),
              (1, None, 'a'),
              (None, None, 'b'),
              (2, True, 'c'))
    ieq(expect, actual)
    ieq(expect, actual)  # verify can iterate twice


def test_fromdicts_3():

    data = [{'foo': 'a', 'bar': 1},
            {'foo': 'b'},
            {'foo': 'c', 'bar': 2, 'baz': True}]
    actual = fromdicts(data, header=['foo', 'bar'])
    expect = (('foo', 'bar'),
              ('a', 1),
              ('b', None),
              ('c', 2))
    ieq(expect, actual)
    ieq(expect, actual)  # verify can iterate twice


def test_fromdicts_onepass():

    # check that fromdicts() only makes a single pass through the data
    data = iter([{'foo': 'a', 'bar': 1},
                 {'foo': 'b', 'bar': 2},
                 {'foo': 'c', 'bar': 2}])
    actual = fromdicts(data, header=['foo', 'bar'])
    expect = (('foo', 'bar'),
              ('a', 1),
              ('b', 2),
              ('c', 2))
    ieq(expect, actual)


def test_fromdicts_ordered():
    data = [OrderedDict([('foo', 'a'), ('bar', 1)]),
            OrderedDict([('foo', 'b')]),
            OrderedDict([('foo', 'c'), ('bar', 2), ('baz', True)])]
    actual = fromdicts(data)
    # N.B., fields come out in original order
    expect = (('foo', 'bar', 'baz'),
              ('a', 1, None),
              ('b', None, None),
              ('c', 2, True))
    ieq(expect, actual)


def test_fromdicts_missing():
    data = [OrderedDict([('foo', 'a'), ('bar', 1)]),
            OrderedDict([('foo', 'b')]),
            OrderedDict([('foo', 'c'), ('bar', 2), ('baz', True)])]
    actual = fromdicts(data, missing="x")
    expect = (('foo', 'bar', 'baz'),
              ('a', 1, "x"),
              ('b', "x", "x"),
              ('c', 2, True))
    ieq(expect, actual)


def test_tojson():

    # exercise function
    table = (('foo', 'bar'),
             ('a', 1),
             ('b', 2),
             ('c', 2))
    f = NamedTemporaryFile(delete=False, mode='r')
    tojson(table, f.name)
    result = json.load(f)
    assert len(result) == 3
    assert result[0]['foo'] == 'a'
    assert result[0]['bar'] == 1
    assert result[1]['foo'] == 'b'
    assert result[1]['bar'] == 2
    assert result[2]['foo'] == 'c'
    assert result[2]['bar'] == 2


def _run_python_with_stdin(code, input_data):
    import subprocess
    import sys

    proc = subprocess.Popen(
        [sys.executable, '-c', code],
        stdin=subprocess.PIPE,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    )
    stdout, stderr = proc.communicate(input_data)
    assert proc.returncode == 0, stderr.decode('utf-8')
    return stdout


def test_tojson_fromcsv_stdin_subprocess():
    csv_data = b'foo,bar\nx,1\ny,2\n'
    code = 'import petl as etl; etl.fromcsv().tojson(sort_keys=True)'

    stdout = _run_python_with_stdin(code, csv_data)

    assert json.loads(stdout.decode('utf-8')) == [
        {'bar': '1', 'foo': 'x'},
        {'bar': '2', 'foo': 'y'},
    ]


def test_tojsonarrays():

    # exercise function
    table = (('foo', 'bar'),
             ('a', 1),
             ('b', 2),
             ('c', 2))
    f = NamedTemporaryFile(delete=False, mode='r')
    tojsonarrays(table, f.name)
    result = json.load(f)
    assert len(result) == 3
    assert result[0][0] == 'a'
    assert result[0][1] == 1
    assert result[1][0] == 'b'
    assert result[1][1] == 2
    assert result[2][0] == 'c'
    assert result[2][1] == 2


def test_tojsonarrays_fromcsv_stdin_subprocess():
    csv_data = b'foo,bar\nx,1\ny,2\n'
    code = 'import petl as etl; etl.fromcsv().tojsonarrays()'

    stdout = _run_python_with_stdin(code, csv_data)

    assert json.loads(stdout.decode('utf-8')) == [
        ['x', '1'],
        ['y', '2'],
    ]


def test_tojsonarrays_header_fromcsv_stdin_subprocess():
    csv_data = b'foo,bar\nx,1\ny,2\n'
    code = (
        'import petl as etl; '
        'etl.fromcsv().tojsonarrays(output_header=True)'
    )

    stdout = _run_python_with_stdin(code, csv_data)

    assert json.loads(stdout.decode('utf-8')) == [
        ['foo', 'bar'],
        ['x', '1'],
        ['y', '2'],
    ]


def test_fromdicts_header_does_not_raise():
    data = [{'foo': 'a', 'bar': 1},
            {'foo': 'b', 'bar': 2},
            {'foo': 'c', 'bar': 2}]
    actual = fromdicts(data)
    assert actual.header()


def test_fromdicts_header_list():
    data = [OrderedDict([('foo', 'a'), ('bar', 1)]),
        OrderedDict([('foo', 'b'), ('bar', 2)]),
        OrderedDict([('foo', 'c'), ('bar', 2)])]
    actual = fromdicts(data)
    header = actual.header()
    assert header == ('foo', 'bar')
    expect = (('foo', 'bar'),
              ('a', 1),
              ('b', 2),
              ('c', 2))
    ieq(expect, actual)
    ieq(expect, actual)


@pytest.fixture
def dicts_generator():
    def generator():
        yield OrderedDict([('foo', 'a'), ('bar', 1)])
        yield OrderedDict([('foo', 'b'), ('bar', 2)])
        yield OrderedDict([('foo', 'c'), ('bar', 2)])
    return generator()


def test_fromdicts_generator_single(dicts_generator):
    actual = fromdicts(dicts_generator)
    expect = (('foo', 'bar'),
              ('a', 1),
              ('b', 2),
              ('c', 2))
    ieq(expect, actual)


def test_fromdicts_generator_twice(dicts_generator):
    actual = fromdicts(dicts_generator)
    expect = (('foo', 'bar'),
              ('a', 1),
              ('b', 2),
              ('c', 2))
    ieq(expect, actual)
    ieq(expect, actual)


def test_fromdicts_generator_header(dicts_generator):
    actual = fromdicts(dicts_generator)
    header = actual.header()
    assert header == ('foo', 'bar')
    expect = (('foo', 'bar'),
              ('a', 1),
              ('b', 2),
              ('c', 2))
    ieq(expect, actual)
    ieq(expect, actual)


def test_fromdicts_generator_random_access():
    def generator():
        for i in range(5):
            yield OrderedDict([('n', i), ('foo', 100*i), ('bar', 200*i)])

    actual = fromdicts(generator(), sample=3)
    assert actual.header() == ('n', 'foo', 'bar')
    # first pass
    it1 = iter(actual)
    first_row1 = next(it1)
    first_row2 = next(it1)
    # second pass
    it2 = iter(actual)
    second_row1 = next(it2)
    second_row2 = next(it2)
    assert first_row1 == second_row1
    assert first_row2 == second_row2
    # reverse order
    second_row3 = next(it2)
    first_row3 = next(it1)
    assert second_row3 == first_row3
    ieq(actual, actual)
    assert actual.header() == ('n', 'foo', 'bar')
    assert len(actual) == 6


def test_fromdicts_generator_missing():
    def generator():
        yield OrderedDict([('foo', 'a'), ('bar', 1)])
        yield OrderedDict([('foo', 'b'), ('bar', 2)])
        yield OrderedDict([('foo', 'c'), ('baz', 2)])
    actual = fromdicts(generator(), missing="x")
    expect = (('foo', 'bar', 'baz'),
              ('a', 1, "x"),
              ('b', 2, "x"),
              ('c', "x", 2))
    ieq(expect, actual)


def test_fromdicts_dicts_method():
    # .dicts() must reach the inherited Table.dicts(), not the input data (#643)
    dummy = dummytable(numrows=3, seed=42)
    data = list(dummy.dicts())
    actual = fromdicts(dummy.dicts())
    assert list(actual.dicts()) == data


def test_fromdicts_dicts_attribute_reads_as_data():
    data = [{'foo': 'a', 'bar': 1}, {'foo': 'b', 'bar': 2}]
    actual = fromdicts(data)
    assert actual.dicts == data
    assert list(actual.dicts) == data
    assert len(actual.dicts) == 2
    assert actual.dicts[1] == data[1]
    assert repr(actual.dicts) == repr(data)


def test_fromdicts_dicts_attribute_assignment():
    actual = fromdicts([{'foo': 'a'}])
    actual.dicts = [{'foo': 'z'}]
    assert list(actual.dicts) == [{'foo': 'z'}]
    ieq((('foo',), ('z',)), actual)


def test_fromdicts_generator_dicts_method(dicts_generator):
    actual = fromdicts(dicts_generator)
    expect = [{'foo': 'a', 'bar': 1},
              {'foo': 'b', 'bar': 2},
              {'foo': 'c', 'bar': 2}]
    assert list(actual.dicts()) == expect


def test_fromdicts_generator_dicts_attribute_is_consumable(dicts_generator):
    # the generator view's .dicts is a one-shot iterator, as it was before
    actual = fromdicts(dicts_generator)
    it = actual.dicts
    assert next(it) == {'foo': 'a', 'bar': 1}
    assert len(list(it)) == 2
    assert list(actual.dicts) == []



# -*- coding: utf-8 -*-
from __future__ import absolute_import, print_function, division

import math

from collections import OrderedDict
from datetime import datetime, date
from decimal import Decimal, localcontext
from tempfile import NamedTemporaryFile

import pytest

from petl.compat import PY3
from petl.transform.basics import cat
from petl.util.base import dicts
from petl.util.vis import look

from petl.test.helpers import ieq

from petl.io.avro import fromavro, toavro, appendavro
from petl.io.avro import _get_definition_from_type_of, precision_and_scale

from petl.test.io.test_avro_schemas import schema0, schema1, schema2, \
    schema3, schema4, schema5, schema6

if PY3:
    from datetime import timezone

try:
    import fastavro
    # import fastavro dependencies
    import pytz
except ImportError as e:
    pytest.skip('SKIP avro tests: %s' % e, allow_module_level=True)
else:
    # region Test Cases

    def test_fromavro11():
        _read_from_mavro_file(table1, schema1)

    def test_fromavro22():
        _read_from_mavro_file(table2, schema2)

    def test_fromavro33():
        _read_from_mavro_file(table3, schema3)

    def test_toavro11():
        _write_to_avro_file(table1, schema1)

    def test_toavro22():
        _write_to_avro_file(table2, schema2)

    def test_toavro33():
        _write_to_avro_file(table3, schema3)

    def test_toavro10():
        _write_to_avro_file(table1, None)

    def test_toavro13():
        _write_to_avro_file(table01, schema0, table10)

    def test_toavro20():
        _write_to_avro_file(table2, None)

    def test_toavro30():
        _write_to_avro_file(table3, None)

    def test_toavro44():
        _write_to_avro_file(table4, schema4)

    def test_toavro55():
        _write_to_avro_file(table5, schema5)

    def test_toavro50():
        _write_to_avro_file(table5, None)

    def test_toavro70():
        _write_to_avro_file(table71, None)

    def test_toavro80():
        _write_to_avro_file(table8, None)

    def test_toavro90():
        _write_to_avro_file(table9, None)

    def test_toavro61():
        _write_to_avro_file(table61, schema6, print_tables=False)

    def test_toavro62():
        _write_to_avro_file(table62, schema6, print_tables=False)

    def test_toavro63():
        _write_to_avro_file(table63, schema6, print_tables=False)

    def test_toavro60():
        _write_to_avro_file(table60, schema6, print_tables=False)

    def test_appendavro11():
        _append_to_avro_file(table11, table12, schema1, table1)

    def test_appendavro22():
        _append_to_avro_file(table21, table22, schema2, table2)

    def test_appendavro10():
        _append_to_avro_file(table11, table12, schema1)

    def test_toavro_troubleshooting10():
        nullable_schema = dict(schema0)
        schema_fields = nullable_schema['fields']
        for field in schema_fields:
            field['type'] = ['null', 'string']
        try:
            _write_temp_avro_file(table1, nullable_schema)
        except ValueError as vex:
            bob = "%s" % vex
            assert 'Bob' in bob
            return
        assert False, 'Failed schema conversion'

    def test_toavro_troubleshooting11():
        table0 = list(table1)
        table0[3][1] = None
        try:
            _write_temp_avro_file(table0, schema1)
        except TypeError as tex:
            joe = "%s" % tex
            assert 'Joe' in joe
            return
        assert False, 'Failed schema conversion'

    # endregion

    # region Decimal schema inference

    def test_toavro_decimal_zero():
        # a zero has no digits to take the logarithm of
        _write_to_avro_file(table_dec_zero, None)
        _assert_decimal_type(table_dec_zero, 8, 2)

    def test_toavro_decimal_small():
        # more fractional digits than the default precision of 8
        _write_to_avro_file(table_dec_small, None)
        _assert_decimal_type(table_dec_small, 9, 9)

    def test_toavro_decimal_integral():
        # positive exponents are trailing zeros, not fractional digits
        _write_to_avro_file(table_dec_integral, None)
        _assert_decimal_type(table_dec_integral, 8, 0)

    def test_toavro_decimal_mixed_scales():
        # writing rescales a value to the scale of the schema, so the
        # precision has to cover the widest unscaled integer of the column
        _write_to_avro_file(table_dec_mixed, None)
        _assert_decimal_type(table_dec_mixed, 20, 11)

    def test_toavro_decimal_nested():
        # arrays and records reach the same inference helper
        _write_to_avro_file(table_dec_nested, None)

    def test_precision_and_scale_of_each_shape():
        for value, prec, scale in decimal_shapes:
            actual = precision_and_scale(value)[:2]
            assert actual == (prec, scale), \
                'precision_and_scale(%s): got %r, expected %r' \
                % (value, actual, (prec, scale))

    def test_inferred_decimal_type_obeys_avro_spec():
        # the spec asks for a precision "greater than zero" and a scale
        # "zero or a positive integer less than or equal to the precision"
        for column in _all_decimal_columns():
            prec, scale = _infer_decimal_type(column)
            assert prec > 0, \
                'column %r got precision %d' % (_labels(column), prec)
            assert 0 <= scale <= prec, \
                'column %r got scale %d > precision %d' \
                % (_labels(column), scale, prec)

    def test_inferred_decimal_type_fits_every_value():
        # a decimal is `unscaled * 10 ** -scale`, so at the scale of the
        # schema each value must stay within the declared precision
        for column in _all_decimal_columns():
            prec, scale = _infer_decimal_type(column)
            for value in column:
                if value is None:
                    continue
                digits = _unscaled_digits(value, scale)
                assert digits <= prec, \
                    'column %r: %s needs %d digits, precision is %d' \
                    % (_labels(column), value, digits, prec)

    def test_precision_and_scale_of_unscaled_value():
        for value, _, _ in decimal_shapes:
            _, scale, bytes_req, unscaled = precision_and_scale(value)
            assert (unscaled < 0) == (value < 0), \
                '%s came back as %d' % (value, unscaled)
            # the defining identity: unscaled * 10**-scale must equal value,
            # trailing zeros included (a plain digit-string compare doesn't
            # catch a positive-exponent value coming back short of them);
            # a wide context avoids scaleb rounding the widest fixtures
            with localcontext() as ctx:
                ctx.prec = 60
                assert Decimal(unscaled).scaleb(-scale) == value, \
                    '%s came back as %d at scale %d' % (value, unscaled, scale)
            assert _significand('{0:f}'.format(value)) \
                == _significand('%d' % unscaled), \
                '%s came back as %d' % (value, unscaled)
            # enough room for the two's complement of the unscaled integer
            assert 2 ** (8 * bytes_req - 1) > abs(unscaled), \
                '%s needs more than %d bytes' % (value, bytes_req)
            assert bytes_req == 1 or 2 ** (8 * bytes_req - 9) <= abs(unscaled), \
                '%s does not need %d bytes' % (value, bytes_req)

    def test_precision_and_scale_of_non_finite():
        for value in [Decimal('NaN'), Decimal('Infinity'), Decimal('-Inf')]:
            try:
                precision_and_scale(value)
            except ValueError as vex:
                # the message has to name the value that cannot be written
                assert str(value) in '%s' % vex, \
                    '%s is not reported by: %s' % (value, vex)
                continue
            assert False, 'built a decimal schema for %s' % value

    # endregion

    # region Execution

    def _read_from_mavro_file(test_rows, test_schema, test_expect=None, print_tables=True):
        _show__expect_rows(test_rows, print_tables)
        test_filename = _create_avro_example(test_schema, test_rows)
        test_actual = fromavro(test_filename)
        test_expect2 = test_rows if test_expect is None else test_expect
        _assert_rows_are_equals(test_expect2, test_actual, print_tables)
        return test_filename

    def _write_temp_avro_file(test_rows, test_schema):
        test_filename = _get_tempfile_path()
        print("Writing avro file:", test_filename)
        toavro(test_rows, test_filename, schema=test_schema)
        return test_filename

    def _write_to_avro_file(test_rows, test_schema, test_expect=None, print_tables=True):
        _show__expect_rows(test_rows, print_tables)
        test_filename = _write_temp_avro_file(test_rows, test_schema)
        test_actual = fromavro(test_filename)
        test_expect2 = test_rows if test_expect is None else test_expect
        _assert_rows_are_equals(test_expect2, test_actual, print_tables)

    def _append_to_avro_file(test_rows1, test_rows2, test_schema, test_expect=None, print_tables=True):
        _show__expect_rows(test_rows1, print_tables)
        _show__expect_rows(test_rows2, print_tables)
        test_filename = _get_tempfile_path()
        toavro(test_rows1, test_filename, schema=test_schema)
        appendavro(test_rows2, test_filename, schema=test_schema)

        test_actual = fromavro(test_filename)
        if test_expect is not None:
            test_expect2 = test_expect
        else:
            test_expect2 = cat(test_rows1, test_rows2)
        _assert_rows_are_equals(test_expect2, test_actual, print_tables)

    # endregion

    # region Helpers

    def _assert_rows_are_equals(test_expect, test_actual, print_tables=True):
        if print_tables:
            _show__rows_from('Actual:', test_actual)
            avro_schema = test_actual.get_avro_schema()
            print('\nSchema:\n', avro_schema)
        ieq(test_expect, test_actual)
        ieq(test_expect, test_actual)  # verify can iterate twice

    def _show__expect_rows(test_rows, print_tables=True, limit=0):
        if print_tables:
            _show__rows_from('\nExpected:', test_rows, limit)

    def _show__rows_from(label, test_rows, limit=0):
        print(label)
        print(look(test_rows, limit=limit))

    def _decs(float_value, rounding=12):
        return Decimal(str(round(float_value, rounding)))

    def _assert_decimal_type(test_rows, prec, scale):
        column = [row[1] for row in test_rows[1:]]
        actual = _infer_decimal_type(column)
        assert actual == (prec, scale), \
            'column %r got %r, expected %r' \
            % (_labels(column), actual, (prec, scale))

    def _all_decimal_columns():
        return [[value] for value, _, _ in decimal_shapes] + decimal_columns

    def _labels(column):
        return [str(value) for value in column]

    def _infer_decimal_type(column):
        '''the precision and scale petl infers for a column of decimals'''
        prev = None
        tdef = None
        for value in column:
            curr, dcurr = _get_definition_from_type_of(u'amount', value, prev)
            # a None row leaves the definition built so far untouched
            if curr is not None:
                tdef = curr
            if dcurr is not None:
                prev = dcurr
        return tdef['precision'], tdef['scale']

    def _significand(text):
        '''the digits of a number, without sign, point or padding zeros'''
        return text.replace('-', '').replace('.', '').strip('0')

    def _unscaled_digits(value, scale):
        '''digit count of `value * 10 ** scale`, taken from its text form

        Deliberately built from the fixed point representation instead of
        `Decimal.as_tuple()`, which is what the code under test reads, and
        without arithmetic, which would round on the decimal context.
        '''
        fixed = '{0:f}'.format(value)
        fraction = fixed.split('.')[1] if '.' in fixed else ''
        assert len(fraction) <= scale, \
            '%s does not fit a scale of %d' % (value, scale)
        text = fixed.replace('-', '').replace('.', '')
        text = text + '0' * (scale - len(fraction))
        return len(text.lstrip('0')) or 1

    def _utc(year, month, day, hour=0, minute=0, second=0, microsecond=0):
        u = datetime(year, month, day, hour, minute, second, microsecond)
        if PY3:
            return u.replace(tzinfo=timezone.utc)
        return u.replace(tzinfo=pytz.utc)

    def _get_tempfile_path(delete_on_close=False):
        f = NamedTemporaryFile(delete=delete_on_close, mode='r')
        test_filename = f.name
        f.close()
        return test_filename

    def _create_avro_example(test_schema, test_table):
        parsed_schema = fastavro.parse_schema(test_schema)
        rows = dicts(test_table)
        with NamedTemporaryFile(delete=False, mode='wb') as fo:
            fastavro.writer(fo, parsed_schema, rows)
            return fo.name

    # endregion

    # region Mockup data

    header1 = [u'name', u'friends', u'age']

    rows1 = [[u'Bob', 42, 33],
             [u'Jim', 13, 69],
             [u'Joe', 86, 17],
             [u'Ted', 23, 51]]

    table1 = [header1] + rows1

    table11 = [header1] + rows1[0:2]
    table12 = [header1] + rows1[2:]

    table01 = [header1[0:2]] + [item[0:2] for item in rows1]
    table10 = [header1] + [item[0:2] + [None] for item in rows1]

    table2 = [[u'name', u'age', u'birthday', u'death', u'insurance', u'deny'],
              [u'pete', 17, date(2012, 10, 11),
                  _utc(2018, 10, 14, 15, 16, 17, 18000), Decimal('1.100'), False],
              [u'mike', 27, date(2002, 11, 12),
                  _utc(2015, 12, 13, 14, 15, 16, 17000), Decimal('1.010'), False],
              [u'zack', 37, date(1992, 12, 13),
                  _utc(2010, 11, 12, 13, 14, 15, 16000), Decimal('123.456'), True],
              [u'gene', 47, date(1982, 12, 25),
                  _utc(2009, 10, 11, 12, 13, 14, 15000), Decimal('-1.010'), False]]

    table21 = table2[0:3]
    table22 = [table2[0]] + table2[3:]

    table3 = [[u'name', u'age', u'birthday', u'death'],
              [u'pete', 17, date(2012, 10, 11),
                  _utc(2018, 10, 14, 15, 16, 17, 18000)],
              [u'mike', 27, date(2002, 11, 12),
                  _utc(2015, 12, 13, 14, 15, 16, 17000)],
              [u'zack', 37, date(1992, 12, 13),
                  _utc(2010, 11, 12, 13, 14, 15, 16000)],
              [u'gene', 47, date(1982, 12, 25),
                  _utc(2009, 10, 11, 12, 13, 14, 15000)]]

    table4 = [[u'name', u'friends', u'age', u'birthday'],
              [u'Bob', 42, 33, date(2012, 10, 11)],
              [u'Jim', 13, 69, None],
              [None, 86, 17, date(1992, 12, 13)],
              [u'Ted', 23, None, date(1982, 12, 25)]]

    table5 = [[u'palette', u'colors'],
              [u'red', [u'red', u'salmon', u'crimson', u'firebrick', u'coral']],
              [u'pink', [u'pink', u'rose']],
              [u'purple', [u'purple', u'violet', u'fuchsia',
                           u'magenta', u'indigo', u'orchid', u'lavender']],
              [u'green', [u'green', u'lime', u'seagreen',
                          u'grass', u'olive', u'forest', u'teal']],
              [u'blue', [u'blue', u'cyan', u'aqua', u'aquamarine',
                         u'turquoise', u'royal', u'sky', u'navy']],
              [u'gold',  [u'gold', u'yellow', u'khaki',
                          u'mocassin', u'papayawhip', u'lemonchiffon']],
              [u'black',  None]]

    header6 = [u'array_string', u'array_record', u'nulable_date',
               u'multi_union_time', u'array_bytes_decimal', u'array_fixed_decimal']

    rows61 = [[u'a', u'b', u'c'],
              [{u'f1': u'1', u'f2': Decimal('654.321')}],
              date(2020, 1, 10),
              _utc(2020, 12, 19, 18, 17, 16, 15000),
              [Decimal('123.456')],
              [Decimal('987.654')], ]

    rows62 = [[u'a', u'b', u'c'],
              [{u'f1': u'1', u'f2': Decimal('654.321')}],
              date(2020, 1, 10),
              _utc(2020, 12, 19, 18, 17, 16, 15000),
              [Decimal('123.456'), Decimal('456.789')],
              [Decimal('987.654'), Decimal('321.123'), Decimal('456.654')]]

    table61 = [header6, rows61]

    table62 = [header6, rows62]

    table63 = [header6, rows61, rows62]

    table60 = [header6, [rows61[0], rows61[1], ]]

    header7 = [u'col', u'sqrt_pow_ij']

    rows70 = [[j, [round(math.sqrt(math.pow(i*j, i+j)), 9)
                   for i in range(1, j+1)]] for j in range(1, 7)]

    rows71 = [[j, [Decimal(str(round(math.sqrt(math.pow(i*j, i+j)), 9)))
                   for i in range(1, j+1)]] for j in range(1, 7)]

    table70 = [header7] + rows70
    table71 = [header7] + rows71

    header8 = [u'number', u'properties']

    rows8 = [[_decs(x), { 
                    u'atan': _decs(math.atan(x)),
                    u'sin': math.sin(x), 
                    u'cos': math.cos(x), 
                    u'tan': math.tan(x), 
                    u'square': x*x, 
                    u'sqrt': math.sqrt(x), 
                    u'log': math.log(x), 
                    u'log10': math.log10(x), 
                    u'exp': math.log10(x), 
                    u'power_x': x**x, 
                    u'power_minus_x': x**-x, 
                }] for x in range(1, 12)]

    table8 = [header8] + rows8

    rows9 = [[1, { u'name': u'Bob', u'age': 20 }],
             [2, { u'name': u'Ted', u'budget': _decs(54321.25) }],
             [2, { u'name': u'Jim', u'color': u'blue' }],
             [2, { u'name': u'Joe', u'alias': u'terminator' }]]

    table9 = [header8] + rows9

    # (value, precision, scale) of the avro decimal logicalType, where the
    # precision is the digit count of the unscaled integer and the scale the
    # count of fractional digits
    decimal_shapes = [
        # zeros, in every shape Decimal can hold one
        (Decimal('0'), 1, 0),
        (Decimal('-0'), 1, 0),
        (Decimal('0.0'), 1, 1),
        (Decimal('0.00'), 2, 2),
        # a zero keeps the width implied by its own exponent
        (Decimal('0E+2'), 3, 0),
        (Decimal('0E-7'), 7, 7),
        # integers, including the exact powers of ten
        (Decimal('1'), 1, 0),
        (Decimal('42'), 2, 0),
        (Decimal('-99'), 2, 0),
        (Decimal('10'), 2, 0),
        (Decimal('100'), 3, 0),
        (Decimal('12345'), 5, 0),
        # ordinary fractions
        (Decimal('1.5'), 2, 1),
        (Decimal('99.99'), 4, 2),
        (Decimal('123.45'), 5, 2),
        (Decimal('100.00'), 5, 2),
        (Decimal('-3.14159'), 6, 5),
        # fewer digits than fractional places
        (Decimal('0.1'), 1, 1),
        (Decimal('0.001'), 3, 3),
        (Decimal('0.00000001'), 8, 8),
        (Decimal('1E-9'), 9, 9),
        (Decimal('1E-12'), 12, 12),
        # positive exponents: trailing zeros are part of the integer
        (Decimal('1E+2'), 3, 0),
        (Decimal('1.5E+3'), 4, 0),
        (Decimal('-2E+5'), 6, 0),
        (Decimal('1E+20'), 21, 0),
        # more digits than the decimal context holds by default
        (Decimal('12345678901234.5678'), 18, 4),
        (Decimal('0.123456789012345678'), 18, 18),
        (Decimal('123456789012345678901234567890'), 30, 0),
    ]

    # columns whose values disagree on scale, precision or both
    decimal_columns = [
        [Decimal('0.00'), Decimal('12.34')],
        [Decimal('1'), Decimal('0'), Decimal('-0'), Decimal('0.00')],
        [Decimal('1.5'), Decimal('0.001')],
        [Decimal('123456789.1'), Decimal('0.12345678901')],
        [Decimal('1E-12'), Decimal('1.5')],
        [Decimal('0.001'), Decimal('1E-12')],
        [Decimal('1E+7'), Decimal('0.5')],
        [Decimal('9.99'), Decimal('1E-9'), Decimal('123456.789')],
        [Decimal('1E+20'), Decimal('0.000001')],
        [Decimal('0.5'), Decimal('1E+2'), Decimal('1E-8')],
        [Decimal('1.100'), None, Decimal('0')],
    ]

    header_dec = [u'name', u'amount']

    table_dec_zero = [header_dec,
                      [u'pete', Decimal('0.00')],
                      [u'mike', Decimal('12.34')],
                      [u'zack', Decimal('-0')]]

    table_dec_small = [header_dec,
                       [u'pete', Decimal('1E-9')],
                       [u'mike', Decimal('0.000000025')]]

    table_dec_integral = [header_dec,
                          [u'pete', Decimal('1E+2')],
                          [u'mike', Decimal('-2E+5')]]

    table_dec_mixed = [header_dec,
                       [u'pete', Decimal('123456789.1')],
                       [u'mike', Decimal('0.12345678901')]]

    table_dec_nested = [[u'name', u'amounts', u'totals'],
                        [u'pete', [Decimal('0.00'), Decimal('1E-9')],
                         OrderedDict([(u'due', Decimal('0')),
                                      (u'paid', Decimal('1E+2'))])]]

    # endregion

    # region testing

    # endregion

# end of tests #



"""
Regression tests for Table subclasses that store a constructor argument
under a name the fluent API already binds on ``Table``. The instance
attribute wins attribute lookup, so the method becomes unreachable and
calling it raises ``TypeError: '...' object is not callable``.

Collisions found so far: ``header`` (issue #555, PR #704), ``dicts``
(issue #643, PR #697), and ``cache``, ``complement``, ``index``, ``skip``.
``test_no_shadowed_api_names`` scans the source for new ones.
"""

from __future__ import absolute_import, print_function, division


import ast
import json as _json
import os
from collections import OrderedDict
from tempfile import NamedTemporaryFile


import pytest


import petl as etl
from petl.compat import PY2
from petl.test.helpers import eq_
from petl.util.base import Table


def _csvfile(text):
    f = NamedTemporaryFile(mode='w', suffix='.csv', delete=False, newline='')
    f.write(text)
    f.close()
    return f.name


def _textfile(text):
    f = NamedTemporaryFile(mode='w', suffix='.txt', delete=False)
    f.write(text)
    f.close()
    return f.name


def _jsonfile(obj):
    f = NamedTemporaryFile(mode='w', suffix='.json', delete=False)
    f.write(_json.dumps(obj))
    f.close()
    return f.name


@pytest.mark.skipif(PY2, reason='NamedTemporaryFile newline kwarg is Python 3 only')
def test_fromcsv_header():
    t = etl.fromcsv(_csvfile('foo,bar\r\n1,2\r\n'))
    eq_(('foo', 'bar'), t.header())


def test_fromtext_header():
    t = etl.fromtext(_textfile('line one\nline two\n'))
    eq_(('lines',), t.header())


def test_fromjson_header():
    t = etl.fromjson(_jsonfile([{'foo': 1, 'bar': 2}]), header=['foo', 'bar'])
    eq_(('foo', 'bar'), t.header())


def test_fromcolumns_header():
    t = etl.fromcolumns([[1, 2], [3, 4]], ['a', 'b'])
    eq_(('a', 'b'), t.header())


def test_cat_header():
    a = etl.wrap([['foo', 'bar'], [1, 2]])
    b = etl.wrap([['foo', 'baz'], [3, 4]])
    t = etl.cat(a, b, header=['foo', 'bar', 'baz'])
    eq_(('foo', 'bar', 'baz'), t.header())


def test_pushheader_header():
    t = etl.pushheader(etl.wrap([[1, 2]]), ['foo', 'bar'])
    eq_(('foo', 'bar'), t.header())


def test_setheader_header():
    t = etl.setheader(etl.wrap([['foo', 'bar'], [1, 2]]), ['a', 'b'])
    eq_(('a', 'b'), t.header())


def test_rowmap_header():
    src = etl.wrap([['foo', 'bar'], [1, 2]])
    t = etl.rowmap(src, lambda row: [row[0]], header=['foo'])
    eq_(('foo',), t.header())


def test_rowmapmany_header():
    src = etl.wrap([['foo', 'bar'], [1, 2]])
    t = etl.rowmapmany(src, lambda row: [[row[0]]], header=['foo'])
    eq_(('foo',), t.header())


def test_rowgroupmap_header():
    src = etl.wrap([['foo', 'bar'], [1, 2]])
    t = etl.rowgroupmap(src, 'foo', lambda k, rows: [[k]], header=['foo'])
    eq_(('foo',), t.header())


def test_rowreduce_header():
    src = etl.wrap([['foo', 'bar'], [1, 2]])
    t = etl.rowreduce(src, 'foo', lambda k, rows: [k], header=['foo'])
    eq_(('foo',), t.header())


def test_mergesort_header():
    a = etl.wrap([['foo', 'bar'], [1, 2]])
    b = etl.wrap([['foo', 'bar'], [3, 4]])
    t = etl.mergesort(a, b, key='foo')
    eq_(('foo', 'bar'), t.header())


def test_validate_header():
    src = etl.wrap([['foo', 'bar'], [1, 2]])
    t = etl.validate(src, header=['foo', 'bar'])
    eq_(('name', 'row', 'field', 'value', 'error'), t.header())


def test_method_matches_function_form():
    # the fluent method must agree with the etl.header() function
    t = etl.fromcolumns([[1, 2]], ['a'])
    eq_(etl.header(t), t.header())


TABLE = [['foo', 'bar'],
         ['a', 1],
         ['b', 2]]

RIGHT = [['foo', 'baz'],
         ['a', True],
         ['b', False]]

SUBSET = [['foo', 'bar'],
          ['b', 2]]


def test_hashjoin_cache():
    t = etl.hashjoin(TABLE, RIGHT, key='foo')
    eq_([('foo', 'bar', 'baz'), ('a', 1, True), ('b', 2, False)],
        list(t.cache()))


def test_hashleftjoin_cache():
    t = etl.hashleftjoin(TABLE, [['foo', 'baz'], ['a', True]], key='foo')
    eq_([('foo', 'bar', 'baz'), ('a', 1, True), ('b', 2, None)],
        list(t.cache()))


def test_hashrightjoin_cache():
    t = etl.hashrightjoin([['foo', 'baz'], ['a', True]], TABLE, key='foo')
    eq_([('foo', 'baz', 'bar'), ('a', True, 1), ('b', None, 2)],
        list(t.cache()))


def test_sort_cache():
    t = etl.sort(TABLE, 'foo', reverse=True)
    eq_([('foo', 'bar'), ('b', 2), ('a', 1)], list(t.cache()))


def test_cache_cache():
    t = etl.wrap(TABLE).cache()
    eq_([('foo', 'bar'), ('a', 1), ('b', 2)],
        [tuple(row) for row in t.cache()])


def test_search_complement():
    t = etl.search(TABLE, 'a|b')
    eq_([('foo', 'bar'), ('a', 1)], list(t.complement(SUBSET)))


def test_rowselect_complement():
    t = etl.select(TABLE, lambda row: True)
    eq_([('foo', 'bar'), ('a', 1)], list(t.complement(SUBSET)))


def test_fieldselect_complement():
    t = etl.select(TABLE, 'foo', lambda v: True)
    eq_([('foo', 'bar'), ('a', 1)], list(t.complement(SUBSET)))


def test_addfield_index():
    t = etl.addfield(TABLE, 'baz', True)
    eq_([('foo', 'bar', 'baz'), ('a', 1, True), ('b', 2, True)], list(t))
    eq_(1, t.index(('a', 1, True)))


def test_movefield_index():
    t = etl.movefield(TABLE, 'bar', 0)
    eq_([('bar', 'foo'), (1, 'a'), (2, 'b')], list(t))
    eq_(1, t.index((1, 'a')))


# fromavro/frombcolz need optional dependencies for a full round trip, so
# feed the readers directly instead of installing fastavro and bcolz here.

class _FakeCtable(object):

    names = ('foo', 'bar')

    def iter(self, outcols=None, skip=0, limit=None):
        return iter([('a', 1), ('b', 2), ('c', 3)][skip:])


def test_fromavro_skip():
    view = etl.fromavro('nonexistent.avro', skips=2)
    assert isinstance(view.skip(1), Table)
    # ordered so the row order does not rely on dict iteration order
    records = [OrderedDict([('foo', v), ('bar', i)])
               for i, v in enumerate(['a', 'b', 'c'], 1)]
    eq_([('c', 3)],
        list(view._read_rows_from(records, ('foo', 'bar'))))


def test_frombcolz_skip():
    view = etl.frombcolz(_FakeCtable(), skip=1)
    eq_([('foo', 'bar'), ('b', 2), ('c', 3)], list(view))
    # Table.skip() drops leading rows, so the next row becomes the header
    eq_([('b', 2), ('c', 3)], list(view.skip(1)))


# The renamed attributes are still read internally, so check the options they
# carry keep working.

def test_hashjoin_cache_option_still_honoured():
    right = etl.wrap([['foo', 'baz'], ['a', True], ['b', False]])
    cached = etl.hashjoin(TABLE, right, key='foo')
    list(cached)
    assert cached.rlookup is not None
    uncached = etl.hashjoin(TABLE, right, key='foo', cache=False)
    list(uncached)
    first = uncached.rlookup
    list(uncached)
    assert uncached.rlookup is not first


def test_sort_cache_option_still_honoured():
    cached = etl.sort(TABLE, 'foo')
    list(cached)
    assert cached._memcache is not None
    uncached = etl.sort(TABLE, 'foo', cache=False)
    list(uncached)
    assert uncached._memcache is None


def test_cacheview_clearcache_still_works():
    t = etl.wrap(TABLE).cache()
    eq_(3, len(list(t)))
    assert t.cachecomplete
    t.clearcache()
    assert not t.cachecomplete
    eq_(3, len(list(t)))


def test_search_complement_option_still_honoured():
    eq_([('foo', 'bar'), ('a', 1)], list(etl.search(TABLE, 'a')))
    eq_([('foo', 'bar'), ('b', 2)],
        list(etl.search(TABLE, 'a', complement=True)))


def test_select_complement_option_still_honoured():
    eq_([('foo', 'bar'), ('b', 2)],
        list(etl.select(TABLE, lambda row: row.foo == 'a', complement=True)))
    eq_([('foo', 'bar'), ('b', 2)],
        list(etl.select(TABLE, 'foo', lambda v: v == 'a', complement=True)))


def test_addfield_index_option_still_honoured():
    eq_([('baz', 'foo', 'bar'), (True, 'a', 1), (True, 'b', 2)],
        list(etl.addfield(TABLE, 'baz', True, index=0)))


def test_movefield_index_option_still_honoured():
    eq_([('bar', 'foo'), (1, 'a'), (2, 'b')],
        list(etl.movefield(TABLE, 'bar', 0)))


def _class_level_names(node):
    """Names bound in a class body, which legitimately override the API."""
    names = set()
    for stmt in node.body:
        if isinstance(stmt, ast.FunctionDef):
            names.add(stmt.name)
        elif isinstance(stmt, ast.Assign):
            for target in stmt.targets:
                if isinstance(target, ast.Name):
                    names.add(target.id)
    return names


def _self_assignments(node):
    for child in ast.walk(node):
        if not isinstance(child, ast.Assign):
            continue
        for target in child.targets:
            if (isinstance(target, ast.Attribute)
                    and isinstance(target.value, ast.Name)
                    and target.value.id == 'self'):
                yield target.attr, child.lineno


def test_no_shadowed_api_names():
    """No Table subclass may store an instance attribute named after a
    method of the fluent API, because the attribute hides the method."""
    api = set(name for name in dir(Table)
              if not name.startswith('__') and callable(getattr(Table, name)))
    assert 'header' in api and 'cache' in api and len(api) > 100

    root = os.path.dirname(os.path.abspath(etl.__file__))
    classes = {}
    for dirpath, dirnames, filenames in os.walk(root):
        if 'test' in dirnames:
            dirnames.remove('test')  # do not scan the suite itself
        for filename in sorted(filenames):
            if not filename.endswith('.py'):
                continue
            path = os.path.join(dirpath, filename)
            # read bytes so the parser applies each module's own coding
            # declaration rather than the platform default encoding
            with open(path, 'rb') as f:
                source = f.read()
            try:
                tree = ast.parse(source)
            except SyntaxError:
                continue  # the py2-only and py3-only modules
            for node in ast.walk(tree):
                if isinstance(node, ast.ClassDef):
                    classes.setdefault(node.name, []).append((path, node))
    assert 'CacheView' in classes and 'SortView' in classes

    def bases(name, seen):
        if name in seen:
            return
        seen.add(name)
        for _, node in classes.get(name, []):
            for base in node.bases:
                if isinstance(base, ast.Name):
                    yield base.id
                    for parent in bases(base.id, seen):
                        yield parent
                elif isinstance(base, ast.Attribute):
                    yield base.attr

    offenders = []
    for name, defs in sorted(classes.items()):
        ancestors = set(bases(name, set()))
        if not ancestors & set(['Table', 'IterContainer']):
            continue
        # a name redefined in the view itself (or in an intermediate view it
        # inherits from) is a deliberate override; the base classes are where
        # the shadowed methods come from, so they do not count
        overridden = set()
        for _, node in defs:
            overridden |= _class_level_names(node)
        for ancestor in ancestors - set(['Table', 'IterContainer']):
            for _, node in classes.get(ancestor, []):
                overridden |= _class_level_names(node)
        for path, node in defs:
            for attr, lineno in _self_assignments(node):
                if attr in api and attr not in overridden:
                    offenders.append('%s:%s: %s.%s shadows Table.%s()'
                                     % (os.path.relpath(path, root), lineno,
                                        name, attr, attr))
    assert not offenders, '\n'.join(offenders)
